# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from ..utils import verify_thicket_structures
from .ttest import __ttest

__statistical_tests = {"ttest": __ttest}


def preference(thicket, columns, comparison_func, test="ttest", *args, **kwargs):
    """Determine a preference between compilers, architecture, platform, etc.

    Designed to take in a thicket and will append eight total columns to the
    aggregated statistics table. As a note, preferred will stand for the preferred
    choice between two options.
        1. columns[0] mean
        2. columns[0] std
        3. columns[1] mean
        4. columns[1] std
        5. columns[0] + columns[1] tvalue
        6. columns[0] + columns[1] tstatistic
        7. columns[0] + columns[1] std preferred
        8. columns[0] + columns[1] mean preferred

    Column names will have the following two formats:
        1. <column_name>_{std,mean}
        2. <column_name> vs <column_name>_{tvalue,tstatistic,preferred}

    Arguments:
        thicket (thicket): Thicket object
        columns (list): List of hardware/timing metrics to determine a preference for.
            Note, if using a columnar joined thicket a list of tuples must be passed in
            with the format (column index, column name). List should be length 2.
        comparison_func (function): User-defined python or lambda function to decide a
            preference.
        test (str): User-selected test.
    """
    if len(columns) != 2:
        raise ValueError("Must specify 2 columns in columns=.")

    if test not in __statistical_tests.keys():
        raise ValueError("Test is not available.")

    verify_thicket_structures(thicket.dataframe, index=["node"], columns=columns)

    if test == "ttest":
        tvalue, t_statistics = __statistical_tests[test](
            thicket, columns, *args, **kwargs
        )

    pref_mean = []
    pref_std = []

    # thicket object without columnar index
    if thicket.dataframe.columns.nlevels == 1:
        for i, t_statistic in enumerate(t_statistics):
            if t_statistic < -1 * tvalue or t_statistic > tvalue:
                pref_mean.append(
                    comparison_func(
                        thicket.statsframe.dataframe[columns[0] + "_mean"][i],
                        thicket.statsframe.dataframe[columns[1] + "_mean"][i],
                    )
                )
                pref_std.append(
                    comparison_func(
                        thicket.statsframe.dataframe[columns[0] + "_std"][i],
                        thicket.statsframe.dataframe[columns[1] + "_std"][i],
                    )
                )
            else:
                pref_mean.append("No preference")
                pref_std.append("No preference")
        aggregated_cols = columns[0] + " vs " + columns[1]
        thicket.statsframe.dataframe[aggregated_cols + "_std_preferred"] = pref_std
        thicket.statsframe.dataframe[aggregated_cols + "_mean_preferred"] = pref_mean
    # columnar joined thicket object
    else:
        idx_mean = [(index, col + "_mean") for index, col in columns]
        idx_std = [(index, col + "_std") for index, col in columns]
        for i, t_statistic in enumerate(t_statistics):
            if t_statistic < -1 * tvalue or t_statistic > tvalue:
                pref_mean.append(
                    comparison_func(
                        thicket.statsframe.dataframe[idx_mean[0]][i],
                        thicket.statsframe.dataframe[idx_mean[1]][i],
                    )
                )
                pref_std.append(
                    comparison_func(
                        thicket.statsframe.dataframe[idx_std[0]][i],
                        thicket.statsframe.dataframe[idx_std[1]][i],
                    )
                )
            else:
                pref_mean.append("No preference")
                pref_std.append("No preference")

        aggregated_cols = (
            str(columns[0]).replace("'", "") + " vs " + str(columns[1]).replace("'", "")
        )
        thicket.statsframe.dataframe[
            (
                "Preference",
                aggregated_cols + "_std_preferred",
            )
        ] = pref_std

        thicket.statsframe.dataframe[
            (
                "Preference",
                aggregated_cols + "_mean_preferred",
            )
        ] = pref_mean



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd
from scipy.stats import ttest_ind_from_stats
from scipy.stats import t

import thicket as th


def __ttest(thicket, columns, alpha=0.05, *args, **kwargs):
    """Perform a ttest on a user-selected thicket and columns.

    Designed to take in a thicket and two columns. For this private function a tvalue
    and list of tstatistics will be returned to preference.py.

    Arguments:
        thicket (thicket): Thicket object
        columns (list): List of hardware/timing metrics to determine a preference for.
            Note, if using a columnar joined thicket a list of tuples must be passed in
            with the format (column index, column name).
        alpha (double): Threshold for statistical significance. Value must be between 0
            and 1. Default is 0.05.

    Returns:
       tvalue (double): Value to be used to determine a preference within preference.py.
       tstatistics (list): List of values to be used to determine a preference within
            preference.py.
    """
    # check to see if alpha value is between 0 and 1
    if alpha <= 0 or alpha >= 1:
        raise ValueError("Value for alpha argument must be between 0 and 1.")

    # check columns contain two columns
    if len(columns) != 2:
        raise ValueError("Columns must be a list of length 2.")

    n = pd.unique(thicket.dataframe.reset_index()["node"])[0]

    # nobs for parameter one for ttest
    nobs_column1 = len(thicket.dataframe.loc[n][columns[0]])
    # nobs for parameter two for ttest
    nobs_column2 = len(thicket.dataframe.loc[n][columns[0]])

    # subtract by len(columns) due to estimating a t-test with two parameters
    # alpha/2 is done for a two tail t-test
    tvalue = t.ppf(q=1 - alpha / 2, df=nobs_column1 + nobs_column2 - len(columns))

    th.mean(thicket, columns)
    th.std(thicket, columns)

    # thicket object with columnar index
    if thicket.dataframe.columns.nlevels > 1:
        mean_columns = [(idx, col + "_mean") for idx, col in columns]
        std_columns = [(idx, col + "_std") for idx, col in columns]
        t_statistics = []
        for i in range(0, len(thicket.statsframe.dataframe)):
            tStatistic = ttest_ind_from_stats(
                mean1=thicket.statsframe.dataframe[mean_columns[0]][i],
                std1=thicket.statsframe.dataframe[std_columns[0]][i],
                nobs1=nobs_column1,
                mean2=thicket.statsframe.dataframe[mean_columns[1]][i],
                std2=thicket.statsframe.dataframe[std_columns[1]][i],
                nobs2=nobs_column2,
                equal_var=False,
            )

            t_statistics.append(tStatistic.statistic)

        # store results into thicket's aggregated statistics table
        aggregated_cols = (
            str(columns[0]).replace("'", "") + " vs " + str(columns[1]).replace("'", "")
        )

        thicket.statsframe.dataframe[
            (
                "Preference",
                aggregated_cols + "_tvalue",
            )
        ] = tvalue
        thicket.statsframe.dataframe[
            (
                "Preference",
                aggregated_cols + "_tstatistic",
            )
        ] = t_statistics

        return tvalue, t_statistics
    # thicket object without columnar index
    else:
        # gather mean and std columns
        mean_columns = [col + "_mean" for col in columns]
        std_columns = [col + "_std" for col in columns]
        t_statistics = []
        for i in range(0, len(thicket.statsframe.dataframe)):
            tStatistic = ttest_ind_from_stats(
                mean1=thicket.statsframe.dataframe[mean_columns[0]][i],
                std1=thicket.statsframe.dataframe[std_columns[0]][i],
                nobs1=nobs_column1,
                mean2=thicket.statsframe.dataframe[mean_columns[1]][i],
                std2=thicket.statsframe.dataframe[std_columns[1]][i],
                nobs2=nobs_column2,
                equal_var=False,
            )

            t_statistics.append(tStatistic.statistic)

        # store results into thicket's aggregated statistics table
        aggregated_cols = columns[0] + " vs " + columns[1]

        thicket.statsframe.dataframe[aggregated_cols + "_tvalue"] = tvalue
        thicket.statsframe.dataframe[aggregated_cols + "_tstatistic"] = t_statistics

        return tvalue, t_statistics



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from collections import OrderedDict

from hatchet import GraphFrame
import numpy as np
import pandas as pd

import thicket.helpers as helpers
from .utils import validate_dataframe, verify_sorted_profile, verify_thicket_structures


class Ensemble:
    """Operations pertaining to ensembling."""

    @staticmethod
    def _unify(thickets, inplace=False):
        """Create union graph from list of thickets and sync their DataFrames.

        Arguments:
            thickets (list): list of Thicket objects
            inplace (bool): whether to modify the original thicket objects or return new

        Returns:
            (tuple): tuple containing:
                (hatchet.Graph): unified graph
                (list): list of Thicket objects
        """
        _thickets = thickets
        if not inplace:
            _thickets = [th.deepcopy() for th in thickets]
        # Unify graphs if "self" and "other" do not have the same graph
        union_graph = _thickets[0].graph
        for i in range(len(_thickets) - 1):
            # Check for same graph id (fast) if not -> check for equality (slow)
            if (
                _thickets[i].graph is _thickets[i + 1].graph
                or _thickets[i].graph == _thickets[i + 1].graph
            ):
                continue
            else:
                union_graph = union_graph.union(_thickets[i + 1].graph)
        for i in range(len(_thickets)):
            # Set all graphs to the union graph
            _thickets[i].graph = union_graph
            # For tree diff. dataframes need to be sorted.
            _thickets[i].dataframe.sort_index(inplace=True)
            # Necessary to change dataframe hatchet id's to match the nodes in the graph
            helpers._sync_nodes_frame(union_graph, _thickets[i].dataframe)
        return union_graph, _thickets

    @staticmethod
    def _columns(
        thickets,
        headers=None,
        metadata_key=None,
    ):
        """Concatenate Thicket attributes horizontally. For DataFrames, this implies expanding
        in the column direction. New column multi-index will be created with columns
        under separate indexer headers.

        Arguments:
            headers (list): List of headers to use for the new columnar multi-index
            metadata_key (str): Name of the column from the metadata tables to replace the 'profile'
                index. If no argument is provided, it is assumed that there is no profile-wise
                relationship between the thickets.

        Returns:
            (Thicket): New ensembled Thicket object
        """

        def _check_structures():
            """Check that the structures of the thicket objects are valid for the incoming operations."""
            # Required/expected format of the data
            for th in thickets:
                verify_thicket_structures(th.dataframe, index=["node", "profile"])
                verify_thicket_structures(th.statsframe.dataframe, index=["node"])
                verify_thicket_structures(th.metadata, index=["profile"])
            # Check for metadata_key in metadata
            if metadata_key:
                for th in thickets:
                    verify_thicket_structures(th.metadata, columns=[metadata_key])
            # Check length of profiles match if metadata key is not provided
            if metadata_key is None:
                for i in range(len(thickets) - 1):
                    if len(thickets[i].profile) != len(thickets[i + 1].profile):
                        raise ValueError(
                            "Length of all thicket profiles must match if 'metadata_key' is not provided. {} != {}".format(
                                len(thickets[i].profile), len(thickets[i + 1].profile)
                            )
                        )
            # Ensure all thickets profiles are sorted. Must be true when metadata_key=None to
            # guarantee performance data table and metadata table match up.
            if metadata_key is None:
                for th in thickets:
                    verify_sorted_profile(th.dataframe)
                    verify_sorted_profile(th.metadata)

        def _create_multiindex_columns(df, upper_idx_name):
            """Helper function to create multi-index column names from a dataframe's
            current columns.

            Arguments:
            df (DataFrame): source dataframe
            upper_idx_name (String): name of the newly added index in the multi-index.
                Prepended before each column as a tuple.

            Returns:
                (list): list of new indicies generated from the source dataframe
            """
            new_idx = []
            for column in df.columns:
                new_tuple = (upper_idx_name, column)
                new_idx.append(new_tuple)
            return new_idx

        def _handle_metadata():
            """Handle operations to create new concatenated columnar axis metadata table."""
            # Update index to reflect performance data table index
            for i in range(len(thickets_cp)):
                thickets_cp[i].metadata.reset_index(drop=True, inplace=True)
            if metadata_key is None:
                for i in range(len(thickets_cp)):
                    thickets_cp[i].metadata.index.set_names("profile", inplace=True)
            else:
                for i in range(len(thickets_cp)):
                    thickets_cp[i].metadata.set_index(metadata_key, inplace=True)
                    thickets_cp[i].metadata.sort_index(inplace=True)

            # Create multi-index columns
            for i in range(len(thickets_cp)):
                thickets_cp[i].metadata.columns = pd.MultiIndex.from_tuples(
                    _create_multiindex_columns(thickets_cp[i].metadata, headers[i])
                )

            # Concat metadata together
            combined_th.metadata = pd.concat(
                [thickets_cp[i].metadata for i in range(len(thickets_cp))],
                axis="columns",
            )

        def _handle_misc():
            """Misceallaneous Thicket object operations."""
            for i in range(1, len(thickets_cp)):
                combined_th.profile += thickets_cp[i].profile  # Update "profile" object
                combined_th.profile_mapping.update(
                    thickets_cp[i].profile_mapping
                )  # Update "profile_mapping" object
            combined_th.profile = [new_mappings[prf] for prf in combined_th.profile]
            profile_mapping_cp = combined_th.profile_mapping.copy()
            for k, v in profile_mapping_cp.items():
                combined_th.profile_mapping[
                    new_mappings[k]
                ] = combined_th.profile_mapping.pop(k)
            combined_th.performance_cols = helpers._get_perf_columns(
                combined_th.dataframe
            )

        def _handle_perfdata():
            """Handle operations to create new concatenated columnar axis performance data table.

            Returns:
                (dict): dictionary mapping old profiles to new profiles
            """
            # Create header list if not provided
            nonlocal headers
            if headers is None:
                headers = [i for i in range(len(thickets))]

            # Update index to reflect performance data table index
            new_mappings = {}  # Dictionary mapping old profiles to new profiles
            if metadata_key is None:  # Create index from scratch
                new_profiles = [i for i in range(len(thickets_cp[0].profile))]
                for i in range(len(thickets_cp)):
                    thickets_cp[i].metadata["new_profiles"] = new_profiles
                    thickets_cp[i].metadata_column_to_perfdata(
                        "new_profiles", drop=True
                    )
                    thickets_cp[i].dataframe.reset_index(level="profile", inplace=True)
                    new_mappings.update(
                        pd.Series(
                            thickets_cp[i]
                            .dataframe["new_profiles"]
                            .map(lambda x: (x, headers[i]))
                            .values,
                            index=thickets_cp[i].dataframe["profile"],
                        ).to_dict()
                    )
                    thickets_cp[i].dataframe.drop("profile", axis=1, inplace=True)
                    thickets_cp[i].dataframe.set_index(
                        "new_profiles", append=True, inplace=True
                    )
                    thickets_cp[i].dataframe.index.rename(
                        "profile", level="new_profiles", inplace=True
                    )
            else:  # Change second-level index to be from metadata's "metadata_key" column
                for i in range(len(thickets_cp)):
                    thickets_cp[i].metadata_column_to_perfdata(metadata_key)
                    thickets_cp[i].dataframe.reset_index(level="profile", inplace=True)
                    new_mappings.update(
                        pd.Series(
                            thickets_cp[i]
                            .dataframe[metadata_key]
                            .map(lambda x: (x, headers[i]))
                            .values,
                            index=thickets_cp[i].dataframe["profile"],
                        ).to_dict()
                    )
                    thickets_cp[i].dataframe.drop("profile", axis=1, inplace=True)
                    thickets_cp[i].dataframe.set_index(
                        metadata_key, append=True, inplace=True
                    )
                    thickets_cp[i].dataframe.sort_index(inplace=True)

            # Create tuple columns
            new_columns = [
                _create_multiindex_columns(th.dataframe, headers[i])
                for i, th in enumerate(thickets_cp)
            ]
            # Clear old metrics (non-tuple)
            combined_th.exc_metrics.clear()
            combined_th.inc_metrics.clear()
            # Update inc/exc metrics
            for i in range(len(new_columns)):
                for col_tuple in new_columns[i]:
                    if col_tuple[1] in thickets_cp[i].exc_metrics:
                        combined_th.exc_metrics.append(col_tuple)
                    if col_tuple[1] in thickets_cp[i].inc_metrics:
                        combined_th.inc_metrics.append(col_tuple)
            # Update columns
            for i in range(len(thickets_cp)):
                thickets_cp[i].dataframe.columns = pd.MultiIndex.from_tuples(
                    new_columns[i]
                )

            # Concat performance data table together
            combined_th.dataframe = pd.concat(
                [thickets_cp[i].dataframe for i in range(len(thickets_cp))],
                axis="columns",
            )

            # Extract "name" columns to upper level
            nodes = list(set(combined_th.dataframe.reset_index()["node"]))
            for node in nodes:
                combined_th.dataframe.loc[node, "name"] = node.frame["name"]
            combined_th.dataframe.drop(
                columns=[(headers[i], "name") for i in range(len(headers))],
                inplace=True,
            )

            # Sort DataFrame
            combined_th.dataframe.sort_index(inplace=True)

            return new_mappings

        def _handle_statsframe():
            """Handle operations to create new concatenated columnar axis aggregated statistics table."""
            # Clear aggregated statistics table
            combined_th.statsframe = GraphFrame(
                graph=combined_th.graph,
                dataframe=helpers._new_statsframe_df(
                    combined_th.dataframe, multiindex=True
                ),
            )

        # Step 0A: Pre-check of data structures
        _check_structures()
        # Step 0B: Variable Initialization
        combined_th = thickets[0].deepcopy()
        thickets_cp = [th.deepcopy() for th in thickets]

        # Step 1: Unify the thickets
        union_graph, _thickets = Ensemble._unify(thickets_cp)
        combined_th.graph = union_graph
        thickets_cp = _thickets

        # Step 2A: Handle performance data tables
        new_mappings = _handle_perfdata()
        # Step 2B: Handle metadata tables
        _handle_metadata()
        # Step 2C: Handle statistics table
        _handle_statsframe()
        # Step 2D: Handle other Thicket objects.
        _handle_misc()

        # Validate dataframe
        validate_dataframe(combined_th.dataframe)

        return combined_th

    @staticmethod
    def _index(thickets, from_statsframes=False):
        """Unify a list of thickets into a single thicket

        Arguments:
            from_statsframes (bool): Whether this method was invoked from from_statsframes

        Returns:
            unify_graph (hatchet.Graph): unified graph,
            unify_df (DataFrame): unified dataframe,
            unify_exc_metrics (list): exclusive metrics,
            unify_inc_metrics (list): inclusive metrics,
            unify_metadata (DataFrame): unified metadata,
            unify_profile (list): profiles,
            unify_profile_mapping (dict): profile mapping
        """

        def _fill_perfdata(perfdata, fill_value=np.nan):
            # Fill missing rows in dataframe with NaN's
            perfdata = perfdata.reindex(
                pd.MultiIndex.from_product(perfdata.index.levels), fill_value=fill_value
            )
            # Replace "NaN" with "None" in columns of string type
            for col in perfdata.columns:
                if pd.api.types.is_string_dtype(perfdata[col].dtype):
                    perfdata[col].replace({fill_value: None}, inplace=True)

            return perfdata

        # Add missing indicies to thickets
        helpers._resolve_missing_indicies(thickets)

        # Initialize attributes
        unify_graph = None
        unify_df = pd.DataFrame()
        unify_inc_metrics = []
        unify_exc_metrics = []
        unify_metadata = pd.DataFrame()
        unify_profile = []
        unify_profile_mapping = OrderedDict()

        # Unification
        unify_graph, thickets = Ensemble._unify(thickets)
        for th in thickets:
            # Extend metrics
            unify_inc_metrics.extend(th.inc_metrics)
            unify_exc_metrics.extend(th.exc_metrics)
            # Extend metadata
            if len(th.metadata) > 0:
                curr_meta = th.metadata.copy()
                unify_metadata = pd.concat([curr_meta, unify_metadata])
            # Extend profile
            if th.profile is not None:
                unify_profile.extend(th.profile)
            # Extend profile mapping
            if th.profile_mapping is not None:
                unify_profile_mapping.update(th.profile_mapping)
            # Extend dataframe
            unify_df = pd.concat([th.dataframe, unify_df])
        # Sort by keys
        unify_profile_mapping = OrderedDict(sorted(unify_profile_mapping.items()))

        # Insert missing rows in dataframe
        unify_df = _fill_perfdata(unify_df)

        # Sort PerfData
        unify_df.sort_index(inplace=True)
        # Sort Metadata
        unify_metadata.sort_index(inplace=True)

        # Remove duplicates in metrics
        unify_inc_metrics = list(set(unify_inc_metrics))
        unify_exc_metrics = list(set(unify_exc_metrics))

        # Validate unify_df
        validate_dataframe(unify_df)

        unify_parts = (
            unify_graph,
            unify_df,
            unify_exc_metrics,
            unify_inc_metrics,
            unify_metadata,
            unify_profile,
            unify_profile_mapping,
        )
        return unify_parts



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd
from scipy import stats

from ..utils import verify_thicket_structures


def check_normality(thicket, columns=None):
    """Determine if the data is normal or non-normal for each node in the performance
    data table.

    Designed to take in a thicket, and append one or more columns to the aggregated
    statistics table. A true boolean value will be appended if the data is normal and a
    false boolean value will be appended if the data is non-normal.

    For this test, the more data the better. Preferably you would want to have 20 data
    points (20 files) in a dataset to have an accurate result.

    Arguments:
        thicket (thicket): Thicket object
        columns (list): List of hardware/timing metrics to perform normality test on.
            Note, if using a columnar joined thicket a list of tuples must be passed in
            with the format (column index, column name).
    """
    if columns is None:
        raise ValueError(
            "To see a list of valid columns, run 'Thicket.performance_cols'."
        )

    verify_thicket_structures(thicket.dataframe, index=["node"], columns=columns)

    # thicket object without columnar index
    if thicket.dataframe.columns.nlevels == 1:
        df = (
            thicket.dataframe.select_dtypes(include="number")
            .reset_index()
            .groupby("node")
            .agg(stats.shapiro)
        )
        for column in columns:
            for i in range(0, len(df[column])):
                pvalue = df[column][i].pvalue

                if pvalue < 0.05:
                    thicket.statsframe.dataframe.loc[
                        df.index[i], column + "_normality"
                    ] = "False"
                elif pvalue > 0.05:
                    thicket.statsframe.dataframe.loc[
                        df.index[i], column + "_normality"
                    ] = "True"
                else:
                    thicket.stataframe.dataframe.loc[
                        df.index[i], column + "_normality"
                    ] = pd.NA
                # check to see if exclusive metric
                if column in thicket.exc_metrics:
                    thicket.statsframe.exc_metrics.append(column + "_normality")
                # check to see if inclusive metric
                else:
                    thicket.statsframe.inc_metrics.append(column + "_normality")
    # columnar joined thicket object
    else:
        df = (
            thicket.dataframe.select_dtypes(include="number")
            .reset_index(level=1)
            .groupby("node")
            .agg(stats.shapiro)
        )
        for idx, column in columns:
            for i in range(0, len(df[(idx, column)])):
                pvalue = df[(idx, column)][i].pvalue

                if pvalue < 0.05:
                    thicket.statsframe.dataframe.loc[
                        df.index[i], (idx, column + "_normality")
                    ] = "False"
                elif pvalue > 0.05:
                    thicket.statsframe.dataframe.loc[
                        df.index[i], (idx, column + "_normality")
                    ] = "True"
                else:
                    thicket.statsframe.dataframe.loc[
                        df.index[i], (idx, column + "_normality")
                    ] = pd.NA
            # check to see if exclusive metric
            if (idx, column) in thicket.exc_metrics:
                thicket.statsframe.exc_metrics.append((idx, column + "_normality"))
            # check to see if inclusive metric
            else:
                thicket.statsframe.inc_metrics.append((idx, column + "_normality"))

        # sort columns in index
        thicket.statsframe.dataframe = thicket.statsframe.dataframe.sort_index(axis=1)



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from ..utils import verify_thicket_structures


def minimum(thicket, columns=None):
    """Determine the minimum for each node in the performance data table.

    Designed to take in a thicket, and append one or more columns to the aggregated
    statistics table for the minimum value for each node.

    The minimum is the lowest observation for a node and its associated profiles.

    Arguments:
        thicket (thicket): Thicket object
        columns (list): List of hardware/timing metrics to determine minimum value for.
            Note, if using a columnar joined thicket a list of tuples must be passed in
            with the format (column index, column name).
    """
    if columns is None:
        raise ValueError(
            "To see a list of valid columns, run 'Thicket.performance_cols'."
        )

    verify_thicket_structures(thicket.dataframe, index=["node"], columns=columns)

    # thicket object without columnar index
    if thicket.dataframe.columns.nlevels == 1:
        df = thicket.dataframe[columns].reset_index().groupby("node").agg(min)
        for column in columns:
            thicket.statsframe.dataframe[column + "_min"] = df[column]
            # check to see if exclusive metric
            if column in thicket.exc_metrics:
                thicket.statsframe.exc_metrics.append(column + "_min")
            # check to see if inclusive metric
            else:
                thicket.statsframe.inc_metrics.append(column + "_min")
    # columnar joined thicket object
    else:
        df = thicket.dataframe[columns].reset_index(level=1).groupby("node").agg(min)
        for idx, column in columns:
            thicket.statsframe.dataframe[(idx, column + "_min")] = df[(idx, column)]
            # check to see if exclusive metric
            if (idx, column) in thicket.exc_metrics:
                thicket.statsframe.exc_metrics.append((idx, column + "_min"))
            # check to see if inclusive metric
            else:
                thicket.statsframe.inc_metrics.append((idx, column + "_min"))

        # sort columns in index
        thicket.statsframe.dataframe = thicket.statsframe.dataframe.sort_index(axis=1)



import warnings

import numpy as np
import pandas as pd
from hatchet.external.console import ConsoleRenderer
from hatchet.util.colormaps import ColorMaps

from ..version import __version__


class ThicketRenderer(ConsoleRenderer):
    # pylint: disable=W1401
    def render_preamble(self):
        lines = [
            r"  _____ _     _      _        _   ",
            r" |_   _| |__ (_) ___| | _____| |_ ",
            r"   | | | '_ \| |/ __| |/ / _ \ __|",
            r"   | | | | | | | (__|   <  __/ |_ ",
            r"   |_| |_| |_|_|\___|_|\_\___|\__|  {:>2}".format("v" + __version__),
            r"",
            r"",
        ]

        return "\n".join(lines)

    def render(self, roots, dataframe, **kwargs):
        self.render_header = kwargs["render_header"]

        if self.render_header:
            result = self.render_preamble()
        else:
            result = ""

        if roots is None:
            result += "The graph is empty.\n\n"
            return result

        self.metric_columns = kwargs["metric_column"]
        self.annotation_column = kwargs["annotation_column"]
        self.precision = kwargs["precision"]
        self.name = kwargs["name_column"]
        self.expand = kwargs["expand_name"]
        self.context = kwargs["context_column"]
        self.rank = kwargs["rank"]
        self.thread = kwargs["thread"]
        self.depth = kwargs["depth"]
        self.highlight = kwargs["highlight_name"]
        self.colormap = kwargs["colormap"]
        self.invert_colormap = kwargs["invert_colormap"]
        self.colormap_annotations = kwargs["colormap_annotations"]
        self.min_value = kwargs["min_value"]
        self.max_value = kwargs["max_value"]

        if self.color:
            self.colors = self.colors_enabled
            # set the colormap based on user input
            self.colors.colormap = ColorMaps().get_colors(
                self.colormap, self.invert_colormap
            )

            if self.annotation_column and self.colormap_annotations:
                self.colors_annotations = self.colors_enabled()
                if isinstance(self.colormap_annotations, (str, list)):
                    if isinstance(self.colormap_annotations, str):
                        self.colors_annotations.colormap = ColorMaps().get_colors(
                            self.colormap_annotations, False
                        )
                    elif isinstance(self.colormap_annotations, list):
                        self.colors_annotations.colormap = self.colormap_annotations
                    self.colors_annotations_mapping = sorted(
                        list(dataframe[self.annotation_column].apply(str).unique())
                    )
                elif isinstance(self.colormap_annotations, dict):
                    self.colors_annotations_mapping = self.colormap_annotations
        else:
            self.colors = self.colors_disabled

        if isinstance(self.metric_columns, (str, tuple)):
            self.primary_metric = self.metric_columns
            self.second_metric = None
        elif isinstance(self.metric_columns, list):
            if len(self.metric_columns) > 2:
                warnings.warn(
                    "More than 2 metrics specified in metric_column=. Tree() will only show 2 metrics at a time. The remaining metrics will not be shown.",
                    SyntaxWarning,
                )
                self.primary_metric = self.metric_columns[0]
                self.second_metric = self.metric_columns[1]
            elif len(self.metric_columns) == 2:
                self.primary_metric = self.metric_columns[0]
                self.second_metric = self.metric_columns[1]
            elif len(self.metric_columns) == 1:
                self.primary_metric = self.metric_columns[0]
                self.second_metric = None

        if self.primary_metric not in dataframe.columns:
            raise KeyError(
                "metric_column={} does not exist in the dataframe, please select a valid column. See a list of the available metrics with GraphFrame.show_metric_columns().".format(
                    self.primary_metric
                )
            )
        if (
            self.second_metric is not None
            and self.second_metric not in dataframe.columns
        ):
            raise KeyError(
                "metric_column={} does not exist in the dataframe, please select a valid column. See a list of the available metrics with GraphFrame.show_metric_columns().".format(
                    self.second_metric
                )
            )

        # grab the min and max value for the primary metric, ignoring inf and
        # nan values

        if "rank" in dataframe.index.names:
            metric_series = (dataframe.xs(self.rank, level=1))[self.primary_metric]
        else:
            metric_series = dataframe[self.primary_metric]
        isfinite_mask = np.isfinite(metric_series.values)
        filtered_series = pd.Series(
            metric_series.values[isfinite_mask], metric_series.index[isfinite_mask]
        )

        self.max_metric = self.max_value if self.max_value else filtered_series.max()
        self.min_metric = self.min_value if self.min_value else filtered_series.min()

        if self.unicode:
            self.lr_arrows = {"◀": "◀ ", "▶": "▶ "}
        else:
            self.lr_arrows = {"◀": "< ", "▶": "> "}

        for root in sorted(roots, key=lambda n: n.frame):
            result += self.render_frame(root, dataframe)

        if self.color is True:
            result += self.render_legend()

        if self.unicode:
            return result
        else:
            return result.encode("utf-8")

    def render_frame(self, node, dataframe, indent="", child_indent=""):
        node_depth = node._depth
        if node_depth < self.depth:
            df_index = node
            node_metric = dataframe.loc[df_index, self.primary_metric]

            metric_precision = "{:." + str(self.precision) + "f}"
            metric_str = (
                self._ansi_color_for_metric(node_metric)
                + metric_precision.format(node_metric)
                + self.colors.end
            )

            if self.second_metric is not None:
                metric_str += " {c.faint}{second_metric:.{precision}f}{c.end}".format(
                    second_metric=dataframe.loc[df_index, self.second_metric],
                    precision=self.precision,
                    c=self.colors,
                )

            if self.annotation_column is not None:
                annotation_content = str(
                    dataframe.loc[df_index, self.annotation_column]
                )
                if self.colormap_annotations:
                    if isinstance(self.colormap_annotations, dict):
                        color_annotation = self.colors_annotations_mapping[
                            annotation_content
                        ]
                    else:
                        color_annotation = self.colors_annotations.colormap[
                            self.colors_annotations_mapping.index(annotation_content)
                            % len(self.colors_annotations.colormap)
                        ]
                    metric_str += " [{}".format(color_annotation)
                    metric_str += "{}".format(annotation_content)
                    metric_str += "{}]".format(self.colors_annotations.end)
                else:
                    metric_str += " [{}]".format(annotation_content)

            if isinstance(dataframe.columns, pd.MultiIndex):
                node_name = dataframe.loc[df_index, (self.name, "")]
            else:
                node_name = dataframe.loc[df_index, self.name]
            if self.expand is False:
                if len(node_name) > 39:
                    node_name = (
                        node_name[:18] + "..." + node_name[(len(node_name) - 18) :]
                    )
            name_str = (
                self._ansi_color_for_name(node_name) + node_name + self.colors.end
            )

            result = "{indent}{metric_str} {name_str}".format(
                indent=indent, metric_str=metric_str, name_str=name_str
            )
            if self.context in dataframe.columns:
                result += " {c.faint}{context}{c.end}\n".format(
                    context=dataframe.loc[df_index, self.context], c=self.colors
                )
            else:
                result += "\n"

            if self.unicode:
                indents = {"├": "├─ ", "│": "│  ", "└": "└─ ", " ": "   "}
            else:
                indents = {"├": "|- ", "│": "|  ", "└": "`- ", " ": "   "}

            # ensures that we never revisit nodes in the case of
            # large complex graphs
            if node not in self.visited:
                self.visited.append(node)
                sorted_children = sorted(node.children, key=lambda n: n.frame)
                if sorted_children:
                    last_child = sorted_children[-1]

                for child in sorted_children:
                    if child is not last_child:
                        c_indent = child_indent + indents["├"]
                        cc_indent = child_indent + indents["│"]
                    else:
                        c_indent = child_indent + indents["└"]
                        cc_indent = child_indent + indents[" "]
                    result += self.render_frame(
                        child, dataframe, indent=c_indent, child_indent=cc_indent
                    )
        else:
            result = ""
            indents = {"├": "", "│": "", "└": "", " ": ""}

        return result



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import copy
import os
import sys
import json
import warnings
from collections import OrderedDict
from hashlib import md5

import pandas as pd
import numpy as np
from hatchet import GraphFrame
from hatchet.query import AbstractQuery, QueryMatcher

from thicket.ensemble import Ensemble
import thicket.helpers as helpers
from .utils import verify_thicket_structures
from .external.console import ThicketRenderer


class Thicket(GraphFrame):
    """Ensemble of profiles, includes a graph and three dataframes, performance data,
    metadata, and aggregated statistics.
    """

    def __init__(
        self,
        graph,
        dataframe,
        exc_metrics=None,
        inc_metrics=None,
        default_metric="time",
        metadata={},
        performance_cols=None,
        profile=None,
        profile_mapping=None,
        statsframe=None,
    ):
        """Create a new thicket from a graph and a dataframe.

        Arguments:
            graph (Graph): graph of nodes in this thicket
            dataframe (DataFrame): pandas DataFrame indexed by Nodes from the graph, and
                potentially other indexes
            exc_metrics (list): list of names of exclusive metrics in the dataframe
            inc_metrics (list): list of names of inclusive metrics in the dataframe
            default_metric (str): primary metric
            metadata (DataFrame): pandas DataFrame indexed by profile hashes, contains
                profile metadata
            performance_cols (list): list of numeric columns within the performance
                dataframe
            profile (list): list of hashed profile strings
            profile_mapping (dict): mapping of hashed profile strings to original strings
            statsframe (DataFrame): pandas DataFrame indexed by Nodes from the graph
        """
        super().__init__(
            graph, dataframe, exc_metrics, inc_metrics, default_metric, metadata
        )
        self.profile = profile
        self.profile_mapping = profile_mapping
        if statsframe is None:
            self.statsframe = GraphFrame(
                graph=self.graph,
                dataframe=helpers._new_statsframe_df(dataframe),
            )
        else:
            self.statsframe = statsframe

        self.performance_cols = helpers._get_perf_columns(self.dataframe)

    def __eq__(self, other):
        """Compare two thicket objects.

        Arguments:
            other (Thicket): Thicket object to compare to

        Returns:
            (bool): True if equal, False otherwise
        """
        assert isinstance(other, Thicket)
        return (
            self.graph == other.graph
            and self.dataframe.equals(other.dataframe)
            and self.exc_metrics == other.exc_metrics
            and self.inc_metrics == other.inc_metrics
            and self.default_metric == other.default_metric
            and self.metadata.equals(other.metadata)
            and self.performance_cols == other.performance_cols
            and self.profile == other.profile
            and self.profile_mapping == other.profile_mapping
            and self.statsframe.graph == other.statsframe.graph
            and self.statsframe.dataframe.equals(other.statsframe.dataframe)
        )

    def __str__(self):
        s = (
            "graph: "
            + helpers._print_graph(self.graph)
            + "\ndataframe:\n"
            + self.dataframe
            + "\nexc_metrics: "
            + self.exc_metrics
            + "\ninc_metrics: "
            + self.inc_metrics
            + "\ndefault_metric: "
            + self.default_metric
            + "\nmetadata:\n"
            + self.metadata
            + "\nperformance_cols:\n"
            + self.performance_cols
            + "\nprofile: "
            + self.profile
            + "\nprofile_mapping: "
            + self.profile_mapping
            + "\nstatsframe:\n"
            + helpers._print_graph(self.statsframe.graph)
            + "\n"
            + self.statsframe.dataframe
        )

        return s

    @staticmethod
    def thicketize_graphframe(gf, prf):
        """Necessary function to handle output from using GraphFrame readers.

        Arguments:
            gf (GraphFrame): hatchet GraphFrame object
            prf (str): profile source of the GraphFrame

        Returns:
            (thicket): Thicket object
        """
        th = Thicket(
            graph=gf.graph,
            dataframe=gf.dataframe,
            exc_metrics=gf.exc_metrics,
            inc_metrics=gf.inc_metrics,
            metadata=gf.metadata,
        )

        if th.profile is None and isinstance(prf, str):
            # Store used profiles and profile mappings using a truncated md5 hash of their string.
            # Resulting int hash will be at least hex_length digits and theoretically up to
            # ceil(log_10(16^n - 1)) digits after conversion.

            # length of the hex string before being converted to an integer.
            hex_length = 8

            hash_arg = int(md5(prf.encode("utf-8")).hexdigest()[:hex_length], 16)
            th.profile = [hash_arg]
            th.profile_mapping = OrderedDict({hash_arg: prf})

            # format metadata as a dict of dicts
            temp_meta = {}
            temp_meta[hash_arg] = th.metadata
            th.metadata = pd.DataFrame.from_dict(temp_meta, orient="index")
            th.metadata.index.set_names("profile", inplace=True)

            # Add profile to dataframe index
            th.dataframe["profile"] = hash_arg
            index_names = list(th.dataframe.index.names)
            index_names.insert(1, "profile")
            th.dataframe.reset_index(inplace=True)
            th.dataframe.set_index(index_names, inplace=True)

        return th

    @staticmethod
    def from_caliper(filename_or_stream, query=None, intersection=False):
        """Read in a Caliper .cali or .json file.

        Arguments:
            filename_or_stream (str or file-like): name of a Caliper output file in
                `.cali` or JSON-split format, or an open file object to read one
            query (str): cali-query in CalQL format
            intersection (bool): whether to perform intersection or union (default)
        """
        return Thicket.reader_dispatch(
            GraphFrame.from_caliper, intersection, filename_or_stream, query
        )

    @staticmethod
    def from_hpctoolkit(dirname, intersection=False):
        """Create a GraphFrame using hatchet's HPCToolkit reader and use its attributes
        to make a new thicket.

        Arguments:
            dirname (str): parent directory of an HPCToolkit experiment.xml file
            intersection (bool): whether to perform intersection or union (default)

        Returns:
            (thicket): new thicket containing HPCToolkit profile data
        """
        return Thicket.reader_dispatch(
            GraphFrame.from_hpctoolkit, intersection, dirname
        )

    @staticmethod
    def from_caliperreader(filename_or_caliperreader, intersection=False):
        """Helper function to read one caliper file.

        Arguments:
            filename_or_caliperreader (str or CaliperReader): name of a Caliper output
                file in `.cali` format, or a CaliperReader object
            intersection (bool): whether to perform intersection or union (default)
        """
        return Thicket.reader_dispatch(
            GraphFrame.from_caliperreader, intersection, filename_or_caliperreader
        )

    @staticmethod
    def reader_dispatch(func, intersection=False, *args, **kwargs):
        """Create a thicket from a list, directory of files, or a single file.

        Arguments:
            func (function): reader function to be used
            intersection (bool): whether to perform intersection or union (default).
            args (list): list of args; args[0] should be an object that can be read from
        """
        ens_list = []
        obj = args[0]  # First arg should be readable object
        extra_args = []
        if len(args) > 1:
            extra_args = args[1:]

        # Parse the input object
        # if a list of files
        if isinstance(obj, (list, tuple)):
            for file in obj:
                ens_list.append(
                    Thicket.thicketize_graphframe(
                        func(file, *extra_args, **kwargs), file
                    )
                )
        # if directory of files
        elif os.path.isdir(obj):
            for file in os.listdir(obj):
                f = os.path.join(obj, file)
                ens_list.append(
                    Thicket.thicketize_graphframe(func(f, *extra_args, **kwargs), f)
                )
        # if single file
        elif os.path.isfile(obj):
            return Thicket.thicketize_graphframe(func(*args, **kwargs), args[0])
        # Error checking
        else:
            if isinstance(obj, str) and not os.path.isfile(obj):
                raise FileNotFoundError("File '" + obj + "' not found.")
            else:
                raise TypeError(
                    "'"
                    + str(type(obj).__name__)
                    + "' is not a valid type to be read from."
                )

        # Perform ensembling operation
        calltree = "union"
        if intersection:
            calltree = "intersection"
        thicket_object = Thicket.concat_thickets(
            thickets=ens_list,
            axis="index",
            calltree=calltree,
        )

        return thicket_object

    @staticmethod
    def concat_thickets(thickets, axis="index", calltree="union", **kwargs):
        """Concatenate thickets together on index or columns. The calltree can either be unioned or
        intersected which will affect the other structures.

        Arguments:
            thickets (list): list of thicket objects
            axis (str): axis to concatenate on -> "index" or "column"
            calltree (str): calltree to use -> "union" or "intersection"

            valid kwargs:
                if axis="index":
                    from_statsframes (bool): Whether this method was invoked from from_statsframes
                if axis="columns":
                    headers (list): List of headers to use for the new columnar multi-index.
                    metadata_key (str): Name of the column from the metadata tables to replace the 'profile'
                index. If no argument is provided, it is assumed that there is no profile-wise
                relationship between the thickets.

        Returns:
            (thicket): concatenated thicket
        """

        def _index(thickets, from_statsframes=False):
            thicket_parts = Ensemble._index(
                thickets=thickets, from_statsframes=from_statsframes
            )

            return Thicket(
                graph=thicket_parts[0],
                dataframe=thicket_parts[1],
                exc_metrics=thicket_parts[2],
                inc_metrics=thicket_parts[3],
                metadata=thicket_parts[4],
                profile=thicket_parts[5],
                profile_mapping=thicket_parts[6],
            )

        def _columns(thickets, headers=None, metadata_key=None):
            combined_thicket = Ensemble._columns(
                thickets=thickets, headers=headers, metadata_key=metadata_key
            )

            return combined_thicket

        if calltree not in ["union", "intersection"]:
            raise ValueError("calltree must be 'union' or 'intersection'")

        if axis == "index":
            ct = _index(thickets, **kwargs)
        elif axis == "columns":
            ct = _columns(thickets, **kwargs)
        else:
            raise ValueError("axis must be 'index' or 'columns'")

        if calltree == "intersection":
            ct = ct.intersection()

        return ct

    @staticmethod
    def columnar_join(thicket_list, header_list=None, metadata_key=None):
        raise ValueError(
            "columnar_join is deprecated. Use 'concat_thickets(axis='columns'...)' instead."
        )

    @staticmethod
    def unify_ensemble(th_list, from_statsframes=False):
        raise ValueError(
            "unify_ensemble is deprecated. Use 'concat_thickets(axis='index'...)' instead."
        )

    @staticmethod
    def from_json(json_thicket):
        # deserialize the json
        thicket_dict = json.loads(json_thicket)
        hatchet_spec = {
            "graph": thicket_dict["graph"],
            "dataframe": thicket_dict["dataframe"],
            "dataframe_indices": thicket_dict["dataframe_indices"],
            "exclusive_metrics": thicket_dict["inclusive_metrics"],
            "inclusive_metrics": thicket_dict["exclusive_metrics"],
        }

        gf = GraphFrame.from_json(json.dumps(hatchet_spec))

        th = Thicket(
            graph=gf.graph,
            dataframe=gf.dataframe,
            exc_metrics=thicket_dict["exclusive_metrics"],
            inc_metrics=thicket_dict["inclusive_metrics"],
            profile=thicket_dict["profile"],
            profile_mapping=thicket_dict["profile_mapping"],
        )

        if "metadata" in thicket_dict:
            mf = pd.DataFrame(thicket_dict["metadata"])
            mf.set_index(mf["profile"], inplace=True)
            if "profile" in mf.columns:
                mf = mf.drop(columns=["profile"])
            th.metadata = mf

        # catch condition where there are no stats
        if "stats" in thicket_dict:
            sf_spec = {
                "graph": thicket_dict["graph"],
                "dataframe": thicket_dict["stats"],
                "dataframe_indices": ["node"],
                "exclusive_metrics": [],
                "inclusive_metrics": [],
            }
            th.statsframe = GraphFrame.from_json(json.dumps(sf_spec))
            th.statsframe.graph = th.graph

        # make and return thicket?
        return th

    def metadata_column_to_perfdata(self, metadata_key, overwrite=False, drop=False):
        """Add a column from the metadata table to the performance data table.

        Arguments:
            metadata_key (str): Name of the column from the metadata table
            overwrite (bool): Determines overriding behavior in performance data table
            drop (bool): Whether to drop the column from the metadata table afterwards
        """
        # Add warning if column already exists in performance data table
        if metadata_key in self.dataframe.columns:
            # Drop column to overwrite, otherwise warn and return
            if overwrite:
                self.dataframe.drop(metadata_key, axis=1, inplace=True)
            else:
                warnings.warn(
                    "Column "
                    + metadata_key
                    + " already exists. Set 'overwrite=True' to force update the column."
                )
                return

        # Add the column to the performance data table
        self.dataframe = self.dataframe.join(
            self.metadata[metadata_key], on=self.dataframe.index.names[1]
        )

        # Drop column
        if drop:
            self.metadata.drop(metadata_key, axis=1, inplace=True)

    def squash(self, update_inc_cols=True):
        """Rewrite the Graph to include only nodes present in the performance
        data table's rows.

        This can be used to simplify the Graph, or to normalize Graph indexes between
        two Thickets.

        Arguments:
            update_inc_cols (boolean, optional): if True, update inclusive columns.

        Returns:
            (Thicket): a newly squashed Thicket object
        """
        squashed_gf = GraphFrame.squash(self, update_inc_cols=update_inc_cols)
        new_graph = squashed_gf.graph
        # The following code updates the performance data and the aggregated statistics
        # table with the remaining (re-indexed) nodes. The dataframe is internally
        # updated in squash(), so we can easily just save it to our thicket performance
        # data. For the aggregated statistics table, we'll have to come up with a better
        # way eventually, but for now, we'll just create a new aggregated statistics
        # table the same way we do when we create a new thicket.
        new_dataframe = squashed_gf.dataframe
        multiindex = False
        if isinstance(self.statsframe.dataframe.columns, pd.MultiIndex):
            multiindex = True
        stats_df = helpers._new_statsframe_df(new_dataframe, multiindex=multiindex)
        sframe = GraphFrame(
            graph=new_graph,
            dataframe=stats_df,
        )
        return Thicket(
            new_graph,
            new_dataframe,
            exc_metrics=self.exc_metrics,
            inc_metrics=self.inc_metrics,
            default_metric=self.default_metric,
            metadata=self.metadata,
            profile=self.profile,
            profile_mapping=self.profile_mapping,
            statsframe=sframe,
        )

    def copy(self):
        """Return a partially shallow copy of the Thicket.

        See GraphFrame.copy() for more details

        Arguments:
            self (Thicket): object to make a copy of

        Returns:
            other (Thicket): copy of self
                (graph ... default_metric): Same behavior as GraphFrame
                metadata (DataFrame): pandas "non-deep" copy of dataframe
                profile (list): copy of self's profile
                profile_mapping (dict): copy of self's profile_mapping
                statsframe (GraphFrame): calls GraphFrame.copy()
        """
        gf = GraphFrame.copy(self)

        return Thicket(
            graph=gf.graph,
            dataframe=gf.dataframe,
            exc_metrics=gf.exc_metrics,
            inc_metrics=gf.inc_metrics,
            default_metric=gf.default_metric,
            metadata=self.metadata.copy(deep=False),
            profile=copy.copy(self.profile),
            profile_mapping=copy.copy(self.profile_mapping),
            statsframe=self.statsframe.copy(),
        )

    def deepcopy(self):
        """Return a deep copy of the Thicket.

        See GraphFrame.deepcopy() for more details

        Arguments:
            self (Thicket): object to make a copy of

        Returns:
            other (Thicket): copy of self
                (graph ... default_metric): same behavior as GraphFrame
                metadata (DataFrame): pandas "deep" copy of dataframe
                profile (list): copy of self's profile
                profile_mapping (dict): copy of self's profile_mapping
                statsframe (GraphFrame): Calls GraphFrame.deepcopy()
        """
        gf = GraphFrame.deepcopy(self)

        return Thicket(
            graph=gf.graph,
            dataframe=gf.dataframe,
            exc_metrics=gf.exc_metrics,
            inc_metrics=gf.inc_metrics,
            default_metric=gf.default_metric,
            metadata=self.metadata.copy(),
            profile=copy.deepcopy(self.profile),
            profile_mapping=copy.deepcopy(self.profile_mapping),
            statsframe=self.statsframe.deepcopy(),
        )

    def tree(
        self,
        metric_column=None,
        annotation_column=None,
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
        colormap_annotations=None,
        render_header=True,
        min_value=None,
        max_value=None,
    ):
        """Visualize the Thicket as a tree

        Arguments:
            metric_column (str, tuple, list, optional): Columns to use the metrics from. Defaults to None.
            annotation_column (str, optional): Column to use as an annotation. Defaults to None.
            precision (int, optional): Precision of shown numbers. Defaults to 3.
            name_column (str, optional): Column of the node name. Defaults to "name".
            expand_name (bool, optional): Limits the lenght of the node name. Defaults to False.
            context_column (str, optional): Shows the file this function was called in (Available with HPCToolkit). Defaults to "file".
            rank (int, optional): Specifies the rank to take the data from. Defaults to 0.
            thread (int, optional): Specifies the thread to take the data from. Defaults to 0.
            depth (int, optional): Sets the maximum depth of the tree. Defaults to 10000.
            highlight_name (bool, optional): Highlights the names of the nodes. Defaults to False.
            colormap (str, optional): Specifies a colormap to use. Defaults to "RdYlGn".
            invert_colormap (bool, optional): Reverts the chosen colormap. Defaults to False.
            colormap_annotations (str, list, dict, optional): Either provide the name of a colormap, a list of colors to use or a dictionary which maps the used annotations to a color. Defaults to None.
            render_header (bool, optional): Shows the Preamble. Defaults to True.
            min_value (int, optional): Overwrites the min value for the coloring legend. Defaults to None.
            max_value (int, optional): Overwrites the max value for the coloring legend. Defaults to None.

        Returns:
            str: String representation of the tree, ready to print
        """
        color = sys.stdout.isatty()
        shell = None
        if metric_column is None:
            metric_column = self.default_metric

        if color is False:
            try:
                import IPython

                shell = IPython.get_ipython().__class__.__name__
            except ImportError:
                pass
            # Test if running in a Jupyter notebook or qtconsole
            if shell == "ZMQInteractiveShell":
                color = True

        if sys.version_info.major == 2:
            unicode = False
        elif sys.version_info.major == 3:
            unicode = True

        return ThicketRenderer(unicode=unicode, color=color).render(
            self.graph.roots,
            self.statsframe.dataframe,
            metric_column=metric_column,
            annotation_column=annotation_column,
            precision=precision,
            name_column=name_column,
            expand_name=expand_name,
            context_column=context_column,
            rank=rank,
            thread=thread,
            depth=depth,
            highlight_name=highlight_name,
            colormap=colormap,
            invert_colormap=invert_colormap,
            colormap_annotations=colormap_annotations,
            render_header=render_header,
            min_value=min_value,
            max_value=max_value,
        )

    @staticmethod
    def from_statsframes(th_list, metadata_key=None):
        """Compose a list of Thickets with data in their statsframes.

        The Thicket's individual aggregated statistics tables are ensembled and become the
        new Thickets performance data table.

        Arguments:
            th_list (list): list of thickets
            metadata_key (str, optional): name of the metadata column to use as
                the new second-level index. Uses the first value so this only makes
                sense if provided column is all equal values and each thicket's columns
                differ in value.

        Returns:
            (thicket): New Thicket object.
        """
        # Pre-check of data structures
        for th in th_list:
            verify_thicket_structures(
                th.dataframe, index=["node", "profile"]
            )  # Required for deepcopy operation
            verify_thicket_structures(
                th.statsframe.dataframe, index=["node"]
            )  # Required for deepcopy operation

        # Setup names list
        th_names = []
        if metadata_key is None:
            for i in range(len(th_list)):
                th_names.append(i)
        else:  # metadata_key was provided.
            for th in th_list:
                # Get name from metadata table
                name_list = th.metadata[metadata_key].tolist()

                if len(name_list) > 1:
                    warnings.warn(
                        f"Multiple values for name {name_list} at thicket.metadata[{metadata_key}]. Only the first will be used."
                    )
                th_names.append(name_list[0])

        th_copy_list = []
        for i in range(len(th_list)):
            th_copy = th_list[i].deepcopy()

            th_id = th_names[i]

            if metadata_key is None:
                idx_name = "profile"
            else:
                idx_name = metadata_key

            # Modify graph
            # Necessary so node ids match up
            th_copy.graph = th_copy.statsframe.graph

            # Modify the performance data table
            df = th_copy.statsframe.dataframe
            df[idx_name] = th_id
            df.set_index(idx_name, inplace=True, append=True)
            th_copy.dataframe = df

            # Adjust profile and profile_mapping
            th_copy.profile = [th_id]
            profile_paths = list(th_copy.profile_mapping.values())
            th_copy.profile_mapping = OrderedDict({th_id: profile_paths})

            # Modify metadata dataframe
            th_copy.metadata[idx_name] = th_id
            th_copy.metadata.set_index(idx_name, inplace=True)

            def _agg_to_set(obj):
                """Aggregate values in 'obj' into a set to remove duplicates."""
                if len(obj) <= 1:
                    return obj
                else:
                    if isinstance(obj.iloc[0], list) or isinstance(obj.iloc[0], set):
                        _set = set(tuple(i) for i in obj)
                    else:
                        _set = set(obj)
                    if len(_set) == 1:
                        return _set.pop()
                    else:
                        return _set

            # Execute aggregation
            th_copy.metadata = th_copy.metadata.groupby(idx_name).agg(_agg_to_set)

            # Append copy to list
            th_copy_list.append(th_copy)

        return Thicket.concat_thickets(th_copy_list, from_statsframes=True)

    def to_json(self, ensemble=True, metadata=True, stats=True):
        jsonified_thicket = {}

        # jsonify graph
        """
        Nodes: {hatchet_nid: {node data, children:[by-id]}}
        """
        formatted_graph_dict = {}

        for n in self.graph.traverse():
            formatted_graph_dict[n._hatchet_nid] = {
                "data": n.frame.attrs,
                "children": [c._hatchet_nid for c in n.children],
                "parents": [c._hatchet_nid for c in n.parents],
            }

        jsonified_thicket["graph"] = [formatted_graph_dict]

        if ensemble:
            jsonified_thicket["dataframe_indices"] = list(self.dataframe.index.names)
            ef = self.dataframe.reset_index()
            ef["node"] = ef["node"].apply(lambda n: n._hatchet_nid)
            jsonified_thicket["dataframe"] = ef.replace({np.nan: None}).to_dict(
                "records"
            )
        if metadata:
            jsonified_thicket["metadata"] = (
                self.metadata.reset_index().replace({np.nan: None}).to_dict("records")
            )
        if stats:
            sf = self.statsframe.dataframe.copy(deep=True)
            sf.index = sf.index.map(lambda n: n._hatchet_nid)
            jsonified_thicket["stats"] = (
                sf.replace({np.nan: None}).reset_index().to_dict("records")
            )

        jsonified_thicket["inclusive_metrics"] = self.inc_metrics
        jsonified_thicket["exclusive_metrics"] = self.exc_metrics
        jsonified_thicket["profile"] = self.profile
        jsonified_thicket["profile_mapping"] = self.profile_mapping

        return json.dumps(jsonified_thicket)

    def intersection(self):
        """Perform an intersection operation on a thicket.

        Nodes not contained in all profiles are removed.

        Returns:
            (thicket): intersected thicket
        """

        # Row that didn't exist will contain "None" in the name column.
        query = (
            QueryMatcher()
            .match(".", lambda row: row["name"].apply(lambda n: n is not None).all())
            .rel("*")
        )
        intersected_th = self.query(query)

        return intersected_th

    def filter_metadata(self, select_function):
        """Filter thicket object based on a metadata key.

        Changes are propagated to the entire thicket object.

        Arguments:
            select_function (lambda function): filter to apply to the metadata table

        Returns:
            (thicket): new thicket object with selected value
        """
        if callable(select_function):
            if not self.metadata.empty:
                # check only 1 index in metadata
                assert self.metadata.index.nlevels == 1

                # Add warning if filtering on multi-index columns
                if isinstance(self.metadata.columns, pd.MultiIndex):
                    warnings.warn(
                        "Filtering on MultiIndex columns will impact the entire row, not just the subsection of the provided MultiIndex."
                    )

                # Get index name
                index_name = self.metadata.index.name

                # create a copy of the thicket object
                new_thicket = self.copy()

                # filter metadata table
                filtered_rows = new_thicket.metadata.apply(select_function, axis=1)
                new_thicket.metadata = new_thicket.metadata[filtered_rows]

                # note index keys to filter performance data table
                index_id = new_thicket.metadata.index.values.tolist()
                # filter performance data table based on the metadata table
                new_thicket.dataframe = new_thicket.dataframe[
                    new_thicket.dataframe.index.get_level_values(index_name).isin(
                        index_id
                    )
                ]

                # create an empty aggregated statistics table with the name column
                new_thicket.statsframe.dataframe = helpers._new_statsframe_df(
                    new_thicket.dataframe
                )
            else:
                raise EmptyMetadataTable(
                    "The provided Thicket object has an empty MetadataTable."
                )

        else:
            raise InvalidFilter("The argument passed to filter must be a callable.")

        return new_thicket

    def filter(self, filter_func):
        """Overloaded generic filter function.

        Provides a helper message for using the thicket filter functions.
        """
        raise RuntimeError(
            "Invalid function: thicket.filter(), please use thicket.filter_metadata() or thicket.filter_stats()"
        )

    def query(self, query_obj, squash=True, update_inc_cols=True):
        """Apply a Hatchet query to the Thicket object.

        Arguments:
            query_obj (AbstractQuery): the query, represented as by a subclass of
                Hatchet's AbstractQuery
            squash (bool): if true, run Thicket.squash before returning the result of
                the query
            update_inc_cols (boolean, optional): if True, update inclusive columns when
                performing squash.

        Returns:
            (Thicket): a new Thicket object containing the data that matches the query
        """
        if isinstance(query_obj, (list, str)):
            raise UnsupportedQuery(
                "Object and String queries from Hatchet are not yet supported in Thicket"
            )
        elif not issubclass(type(query_obj), AbstractQuery):
            raise TypeError(
                "Input to 'query' must be a Hatchet query (i.e., list, str, or subclass of AbstractQuery)"
            )
        dframe_copy = self.dataframe.copy()
        index_names = self.dataframe.index.names
        dframe_copy.reset_index(inplace=True)
        query = query_obj
        # TODO Add a conditional here to parse Object and String queries when supported
        query_matches = query.apply(self)
        filtered_df = dframe_copy.loc[dframe_copy["node"].isin(query_matches)]
        if filtered_df.shape[0] == 0:
            raise EmptyQuery("The provided query would have produced an empty Thicket.")
        filtered_df.set_index(index_names, inplace=True)

        filtered_th = self.deepcopy()
        filtered_th.dataframe = filtered_df

        if squash:
            return filtered_th.squash(update_inc_cols=update_inc_cols)
        return filtered_th

    def groupby(self, by):
        """Create sub-thickets based on unique values in metadata column(s).

        Arguments:
            by (mapping, function, label, pd.Grouper or list of such): Used to determine the groups for the groupby. See pandas.DataFrame.groupby() for more details.

        Returns:
            (list): list of (sub)thickets
        """
        if not self.metadata.empty:
            # group metadata table by unique values in a column
            sub_metadataframes = self.metadata.groupby(by, dropna=False)

            # dictionary of sub_thickets
            sub_thickets = {}

            # for all unique groups of metadata table
            for key, df in sub_metadataframes:
                # create a thicket copy
                sub_thicket = self.deepcopy()

                # return unique group as the metadata table
                sub_thicket.metadata = df

                # find profiles in current unique group and filter performance data
                # table
                profile_id = df.index.values.tolist()
                sub_thicket.dataframe = sub_thicket.dataframe[
                    sub_thicket.dataframe.index.get_level_values("profile").isin(
                        profile_id
                    )
                ]

                # Updates the profiles to only contain the remaining ones
                profile_mapping_tmp = sub_thicket.profile_mapping.copy()
                for profile_mapping_key in profile_mapping_tmp:
                    if profile_mapping_key not in profile_id:
                        sub_thicket.profile_mapping.pop(profile_mapping_key)

                sub_thicket.profile = [
                    profile for profile in sub_thicket.profile if profile in profile_id
                ]

                # clear the aggregated statistics table for current unique group
                sub_thicket.statsframe.dataframe = helpers._new_statsframe_df(
                    sub_thicket.dataframe
                )

                # add thicket to dictionary
                sub_thickets[key] = sub_thicket
        else:
            raise EmptyMetadataTable(
                "The provided Thicket object has an empty metadata table."
            )

        print(len(sub_thickets), " thickets created...")
        print(sub_thickets)

        return sub_thickets

    def filter_stats(self, filter_function):
        """Filter thicket object based on a stats column.

        Propagate changes to the entire thicket object.

        Arguments:
            select_function (lambda function): filter to apply to the aggregated
                statistics table

        Returns:
            (thicket): new thicket object with applied filter function
        """
        # copy thicket
        new_thicket = self.copy()

        # filter aggregated statistics table based on greater than restriction
        filtered_rows = new_thicket.statsframe.dataframe.apply(filter_function, axis=1)
        new_thicket.statsframe.dataframe = new_thicket.statsframe.dataframe[
            filtered_rows
        ]

        # filter performance data table based on filtered nodes
        filtered_nodes = new_thicket.statsframe.dataframe.index.values.tolist()
        new_thicket.dataframe = new_thicket.dataframe[
            new_thicket.dataframe.index.get_level_values("node").isin(filtered_nodes)
        ]

        # filter nodes in the graphframe based on the dataframe nodes
        # TODO see if the new Thicket.squash function will work here
        filtered_graphframe = GraphFrame.squash(new_thicket)
        new_thicket.graph = filtered_graphframe.graph
        new_thicket.statsframe.graph = filtered_graphframe.graph

        return new_thicket

    def get_unique_metadata(self):
        """Get unique values per column in metadata.

        Returns:
            (dict): alphabetical ordered dictionary with key's being the column names
                and the values being unique values for a metadata column.
        """
        unique_meta = {}

        # thicket object without columnar index
        if self.dataframe.columns.nlevels == 1:
            for col in self.metadata.columns:
                # skip columns where the values are a list
                if isinstance(self.metadata[col].iloc[0], list):
                    continue
                else:
                    unique_entries = self.metadata[col].unique().tolist()
                    unique_meta[col] = unique_entries

            sorted_meta = dict(sorted(unique_meta.items(), key=lambda x: x[0].lower()))
        # columnar joined thicket object
        else:
            sorted_meta = []
            for idx in list(self.metadata.columns.levels[0]):
                for col in self.metadata[idx].columns:
                    if isinstance(self.metadata[idx][col].iloc[0], list):
                        continue
                    else:
                        unique_entries = self.metadata[idx][col].unique().tolist()
                        unique_meta[col] = unique_entries

                sorted_meta.append(
                    (idx, dict(sorted(unique_meta.items(), key=lambda x: x[0].lower())))
                )

        return sorted_meta


class InvalidFilter(Exception):
    """Raised when an invalid argument is passed to the filter function."""


class EmptyMetadataTable(Exception):
    """Raised when a Thicket object argument is passed with an empty MetadataTable to
    the filter function.
    """


class UnsupportedQuery(Exception):
    """Raised when an object query or string query are provided to the 'query' function
    because those types of queries are not yet supported in Thicket.
    """


class EmptyQuery(Exception):
    """Raised when a query would result in an empty Thicket object."""



# Copyright 2022 Lawrence Livermore National Security, LLC and other
# Thicket Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from collections import OrderedDict


def validate_dataframe(df):
    """Check validity of a Thicket DataFrame."""

    def _check_duplicate_inner_idx(df):
        """Check for duplicate values in the innermost index."""
        for node in set(df.index.get_level_values("node")):
            inner_idx_values = sorted(
                df.loc[node].index.get_level_values(df.index.nlevels - 2).tolist()
            )
            inner_idx_values_set = sorted(list(set(inner_idx_values)))
            if inner_idx_values != inner_idx_values_set:
                raise IndexError(
                    f"The Thicket.dataframe's index has duplicate values. {inner_idx_values}"
                )

    def _check_missing_hnid(df):
        """Check if there are missing hatchet nid's."""
        i = 0
        set_of_nodes = set(df.index.get_level_values("node"))
        for node in set_of_nodes:
            if hash(node) != i:
                raise ValueError(
                    f"The Thicket.dataframe's index is either not sorted or has a missing node. {hash(node)} ({node}) != {i}"
                )
            i += 1

    def _validate_name_column(df):
        """Check if all of the values in a node's name column are either its name or None."""
        for node in set(df.index.get_level_values("node")):
            names = set(df.loc[node]["name"].tolist())
            node_name = node.frame["name"]
            for name in names:
                if name != node_name and name is not None:
                    raise ValueError(
                        f"Value in the Thicket.dataframe's 'name' column is not valid. {name} != {node_name}"
                    )

    _check_duplicate_inner_idx(df)
    _check_missing_hnid(df)
    _validate_name_column(df)


def verify_sorted_profile(thicket_component):
    """Assertion to check if profiles are sorted in a thicket dataframe

    Arguments:
        thicket_component (DataFrame): component of thicket to check
    """
    profile_index_values = list(
        OrderedDict.fromkeys(thicket_component.index.get_level_values("profile"))
    )
    if profile_index_values != sorted(profile_index_values):
        raise ValueError(
            "The profiles in this dataframe must be sorted. Try 'pandas.DataFrame.sort_index'."
        )


def verify_thicket_structures(thicket_component, columns=[], index=[]):
    """Assertion for missing input requirements to execute thicket functions.

    Arguments:
        thicket_components (DataFrame): component of thicket to check
        columns (list): list of columns to check for in the component
        index (list): list of index levels to check for in the component

    Returns:
        Raises an error if any columns or index levels are not in component, continues
            program if all columns and index levels are in component
    """
    if not isinstance(columns, list):
        raise RuntimeError("columns= must be specified as a list")
    if not isinstance(index, list):
        raise RuntimeError("index= must be specified as a list")

    # collect component columns and index
    component_columns = thicket_component.columns.tolist()
    component_index = thicket_component.index.names

    # check existence of columns and index
    column_result = all(elem in component_columns for elem in columns)
    index_result = all(elem in component_index for elem in index)

    # store missing columns or index if they exist
    missing_columns = list(set(columns).difference(component_columns))
    missing_index = list(set(index).difference(component_index))

    if not column_result and not index_result:
        raise RuntimeError(
            "Specified column(s) not found: "
            + str(missing_columns)
            + "\nMissing index level(s): "
            + str(missing_index)
            + " required for the function"
        )
    elif not column_result and index_result:
        raise RuntimeError("Specified column(s) not found: " + str(missing_columns))
    elif column_result and not index_result:
        raise RuntimeError(
            "Missing index level(s): "
            + str(missing_index)
            + " required for the function"
        )
